Intlr <- function(exp, obs, target_grid, int_method, lr_method, predictors = NULL, loocv = FALSE, ncores = 1) {

  require(multiApply) 

  #-----------------------------------
  # Checkings
  #-----------------------------------
  # input exp and obs must be s2dv_cube objects
  if (!inherits(exp,'s2dv_cube')) {
    stop("Parameter 'exp' must be of the class 's2dv_cube'")
  }
 
  # input exp and obs must be s2dv_cube objects
  if (!inherits(obs,'s2dv_cube')) {
    stop("Parameter 'obs' must be of the class 's2dv_cube'")
  }

  # sdate must be the time dimension in the input data
  stopifnot('sdate' %in% names(dim(exp$data)))
  stopifnot('sdate' %in% names(dim(obs$data))) 
  
  # the parametre 'predictors' must contain the same sdates as exp$data
  if (is.array(predictors)) {
    stopifnot('sdate' %in% names(dim(predictors)))
    stopifnot(dim(predictors)['sdate'] == dim(exp$data)['sdate'])
  }

  #-----------------------------------
  # Interpolation
  #-----------------------------------
  # Interpolate coarse-scale data 
  lonmin <- attr(exp$lon,"first_lon") 
  lonmax <- attr(exp$lon,"last_lon")
  latmin <- attr(exp$lat,"first_lat")
  latmax <- attr(exp$lat,"last_lat")
  exp_interpolated <- Interpolation(exp = exp, target_grid = target_grid, method = int_method,
                                    crop = c(lonmin, lonmax, latmin, latmax))

  # Interpolate obs to the same target grid to ensure the matching with exp coordinates
  obs_interpolated <- Interpolation(exp = obs, target_grid = target_grid, method = int_method,
                                    crop = c(lonmin, lonmax, latmin, latmax))

  # Ensure that the interpolated exp and obs coordinates match
  .check_coords(lat1 = exp_interpolated$lat, lon1 = exp_interpolated$lon, 
                lat2 = obs_interpolated$lat, lon2 = obs_interpolated$lon)

  #-----------------------------------
  # Linear regressions
  #-----------------------------------
  # Pointwise linear regression
  # Predictor: model data
  # Predictand: observations
  if (lr_method == 'basic') {
    predictor <- exp_interpolated$data
    predictand <- obs_interpolated$data
    target_dims_predictor <- 'sdate'
    target_dims_predictand <- 'sdate'
   } 

  # (Multi) linear regression with large-scale predictors
  # Predictor: passed through the parameter 'predictors' by the user. Can be model or observations
  # Predictand: model data
  else if (lr_method == 'large-scale') {  
    if (is.null(predictors)) {
      stop("The large-scale predictors must be passed through the parametre 'predictors'")
    }  
    predictand <- exp_interpolated$data
    predictor <- predictors
    
    var_dim_in_predictor <- names(dim(predictors))[names(dim(predictors)) != 'sdate']
    target_dims_predictor <- c('sdate', var_dim_in_predictor)
    target_dims_predictand <- 'sdate'
  } 

  # Multi-linear regression with the four nearest neighbours
  # Predictors: model data
  # Predictand: observations
  else if (lr_method == '4nn') {
    predictor <- Apply(list(exp_interpolated$data), target_dims = list(c('lat','lon')), fun = find_4nn)$output1
    predictand <- obs_interpolated$data
    target_dims_predictor <- c('sdate','nn')
    target_dims_predictand <- 'sdate'
  }

  else {
    stop(paste0(lr_method, " method is not implemented yet"))
  }

  # Apply the linear regressions
  res <- Apply(list(predictor, predictand), target_dims = list(target_dims_predictor, target_dims_predictand), 
               fun = .intlr, loocv = loocv, ncores = ncores)$output1
 
  names(dim(res))[1] <- 'sdate'

  # Reorder dimensions to match those of the input model data
  res <- .reorder_dims(arr_ref = exp$data, arr_to_reorder = res)

  #-----------------------------------
  # Create an s2dv_cube object
  #-----------------------------------
  res_s2dv <- suppressWarnings(s2dv_cube(data = res, lon = exp_interpolated$lon, lat = exp_interpolated$lat))

  return(res_s2dv)
}

#-----------------------------------
# Atomic function to generate and apply the linear regressions
#-----------------------------------
.intlr <- function(x, y, loocv) {

  tmp_df <- data.frame(x = x, y = y)

  # training
  lm1 <- train_lm(df = tmp_df, loocv = loocv)  

  # prediction
  res <- pred_lm(lm1 = lm1, df = tmp_df, loocv = loocv)

  return(res)
 
}

#-----------------------------------
# Function to generate the linear regressions.
# Returns a list
#-----------------------------------
train_lm <- function(df, loocv) {

  # Remove columns containing only NA's
  df <- df[ , apply(df, 2, function(x) !all(is.na(x)))]

  if (loocv) {
   
    lm1 <- lapply(1:nrow(df), function(j) lm(df[-j,], formula = y ~ .))
 
  } else {

    lm1 <- list(lm(data = df, formula = y ~ .))
  }

  return(lm1)
}

#-----------------------------------
# Function to apply the linear regressions.
#-----------------------------------
pred_lm <- function(df, lm1, loocv) {

  if (loocv) {

    pred_vals <- sapply(1:nrow(df), function(j) predict(lm1[[j]], df[j,]))
  
  } else {
  
    pred_vals_ls <- lapply(lm1, predict, data = df)    
    pred_vals <- unlist(pred_vals_ls)
  } 
 
  return(pred_vals)
}

#-----------------------------------
# Function to find the four nearest neighbours.
# Notes the case for the latitude and longitude borders
#-----------------------------------
find_4nn <- function(object) {

  nearests <- InsertDim(object, posdim = 3, lendim = 4, name = 'nn')
  
  for (i in seq(dim(object)['lon'])) {
   lon_max_border <- FALSE
   lon_min_border <- FALSE 
   if (((i + 1) > dim(object)['lon'])) {
      lon_max_border <- TRUE
   } else if (((i - 1) < 1)) {
      lon_min_border <- TRUE
   }
    for (j in seq(dim(object)['lat'])) {
      lat_max_border <- FALSE
      lat_min_border <- FALSE
      if (((j + 1) > dim(object)['lat'])) {
        lat_max_border <- TRUE
      } else if (((j - 1) < 1)) {
        lat_min_border <- TRUE
      }
      if (lon_max_border) {
        nearests[j, i, 2] <- NA
        nearests[j, i, 4] <- object[j, i - 1]
      } else if (lon_min_border) {
        nearests[j, i, 2] <- object[j, i + 1]
        nearests[j, i, 4] <- NA
      } else {
        nearests[j, i, 2] <- object[j, i + 1]
        nearests[j, i, 4] <- object[j, i - 1]
      }
      if (lat_max_border) {
        nearests[j, i, 1] <- NA
        nearests[j, i, 3] <- object[j - 1, i]
      } else if (lat_min_border) {
        nearests[j, i, 1] <- object[j + 1, i]
        nearests[j, i, 3] <- NA
      } else {
        nearests[j, i, 1] <- object[j + 1, i]
        nearests[j, i, 3] <- object[j - 1, i]
      }
    }
  }

  return(nearests)
}


